#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>

#include "utils_graphics.h"


#define MAX_UNITS           2.0f
#define SCREEN_WIDTH        640
#define SCREEN_HEIGHT       480

#define POSITION            0
#define TEX_COORD           1

SDL_Window* gWindow = NULL;
SDL_GLContext gContext;
SDL_Surface *image2d;
float x_angle;
float y_angle;
float z_angle;
unsigned int last_time;

GLuint vertex_shader;
GLuint fragment_shader;
GLuint shader_program;
GLuint a_position_loc;
GLuint a_texture_coords_loc;
GLuint u_mvpMatrix_loc;

GLuint vertex_array;
GLuint vertex_buffer;
GLuint vertex_indices_buffer;
GLuint texture_coords_buffer;

float projection_matrix[16];
float modelview_matrix[16];
float mvp_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];

GLfloat left = -MAX_UNITS;
GLfloat right = MAX_UNITS;
GLfloat bottom = -MAX_UNITS;
GLfloat top = MAX_UNITS;
GLfloat near = MAX_UNITS + 1.0f;
GLfloat far = MAX_UNITS + 5.0f;

int texture;

GLfloat cube[24*3] = {
      // Front face
      -1.0, -1.0,  1.0,
       1.0, -1.0,  1.0,
       1.0,  1.0,  1.0,
      -1.0,  1.0,  1.0,

      // Back face
      -1.0, -1.0, -1.0,
      -1.0,  1.0, -1.0,
       1.0,  1.0, -1.0,
       1.0, -1.0, -1.0,

      // Top face
      -1.0,  1.0, -1.0,
      -1.0,  1.0,  1.0,
       1.0,  1.0,  1.0,
       1.0,  1.0, -1.0,

      // Bottom face
      -1.0, -1.0, -1.0,
       1.0, -1.0, -1.0,
       1.0, -1.0,  1.0,
      -1.0, -1.0,  1.0,

      // Right face
       1.0, -1.0, -1.0,
       1.0,  1.0, -1.0,
       1.0,  1.0,  1.0,
       1.0, -1.0,  1.0,

      // Left face
      -1.0, -1.0, -1.0,
      -1.0, -1.0,  1.0,
      -1.0,  1.0,  1.0,
      -1.0,  1.0, -1.0
};

GLuint cube_indices[36] = {
      0, 1, 2,      0, 2, 3,    // Front face
      4, 5, 6,      4, 6, 7,    // Back face
      8, 9, 10,     8, 10, 11,  // Top face
      12, 13, 14,   12, 14, 15, // Bottom face
      16, 17, 18,   16, 18, 19, // Right face
      20, 21, 22,   20, 22, 23  // Left face
};

GLfloat tex_coords[24*2] = {
      // Front face
      0.0, 1.0,
      1.0, 1.0,
      1.0, 0.0,
      0.0, 0.0,

      // Back face
      1.0, 1.0,
      1.0, 0.0,
      0.0, 0.0,
      0.0, 1.0,

      // Top face
      0.0, 0.0,
      0.0, 1.0,
      1.0, 1.0,
      1.0, 0.0,

      // Bottom face
      1.0, 0.0,
      0.0, 0.0,
      0.0, 1.0,
      1.0, 1.0,

      // Right face
      1.0, 1.0,
      1.0, 0.0,
      0.0, 0.0,
      0.0, 1.0,

      // Left face
      0.0, 1.0,
      1.0, 1.0,
      1.0, 0.0,
      0.0, 0.0,
};




int init();
int initGL();
int check_gl_error();
void cleanup();
int set_shaders();
int init_buffers();
void update();
void render();

char *text_file_read(char *fname);
void print_shader_info_log(GLuint obj);
void print_program_info_log(GLuint obj);




int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    gWindow = SDL_CreateWindow("Kostka Progmar", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (gWindow == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    gContext = SDL_GL_CreateContext(gWindow);
    if (gContext == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    int imgFlags = IMG_INIT_PNG;
    if (!(IMG_Init(imgFlags) & imgFlags)) {
        printf("Błąd podczas inicjalizacji biblioteki SDL_image: %s\n", IMG_GetError());
        return 0;
    }

    char *path = "lkprogmar.png";
    image2d = IMG_Load(path);
    if (image2d == NULL) {
        printf("Błąd podczas ładowania pliku tekstury %s: %s\n", path, SDL_GetError());
        return 0;
    }

    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    x_angle = 0.0f;
    y_angle = 0.0f;
    z_angle = 0.0f;

    last_time = SDL_GetTicks();
    return 1;
}

int initGL() {
    GLenum err = GL_NO_ERROR;
    GLfloat fovy = 45.0;
    GLfloat aspect = 1.0f;
    unsigned int width = SCREEN_WIDTH;
    unsigned int height = SCREEN_HEIGHT;

    glewExperimental = GL_TRUE;
    err = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (err != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(err));
        return 0;
    }

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glEnable(GL_DEPTH_TEST);

    //printf("surface->format->BytesPerPixel: %d\n", image2d->format->BytesPerPixel);
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, image2d->w, image2d->h, GL_RGBA, GL_UNSIGNED_BYTE, image2d->pixels);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    if (image2d) {
        SDL_FreeSurface(image2d);
        image2d = NULL;
    }
    glBindTexture(GL_TEXTURE_2D, 0);


    glViewport(0, 0, width, height);

    if (height > 0) {
        aspect = (float)width / (float)height;
    }

    get_perspective_matrix(projection_matrix, fovy, aspect, near, far);

    if (!set_shaders()) {
        cleanup();
        return -1;
    }
    if (!init_buffers()) {
        cleanup();
        return -1;
    }

    return 1;
}


int set_shaders() {
    char *shader_source = NULL;
    GLint status;

    shader_source = text_file_read("kostka_progmar.vshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera wierzchołków\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    vertex_shader = glCreateShader(GL_VERTEX_SHADER);
    //printf("vertex_shader: %d\n", vertex_shader);
    glShaderSource(vertex_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(vertex_shader);
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera wierzchołków\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_source = text_file_read("kostka_progmar.fshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera fragmentów\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
    //printf("fragment_shader: %d\n", fragment_shader);
    glShaderSource(fragment_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(fragment_shader);
    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera fragmentów\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_program = glCreateProgram();
    glAttachShader(shader_program, vertex_shader);
    glAttachShader(shader_program, fragment_shader);
    glLinkProgram(shader_program);
    glGetProgramiv(shader_program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna konsolidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }
    glValidateProgram(shader_program);
    glGetProgramiv(shader_program, GL_VALIDATE_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna walidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }

    glUseProgram(shader_program);

    return 1;
}


int init_buffers() {
    glGenVertexArrays(1, &vertex_array);
    glBindVertexArray(vertex_array);

    glGenBuffers(1, &vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(cube), cube, GL_STATIC_DRAW);
    a_position_loc = glGetAttribLocation(shader_program, "in_position");
    glVertexAttribPointer(a_position_loc, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(a_position_loc);

    glGenBuffers(1, &vertex_indices_buffer);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vertex_indices_buffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(cube_indices), cube_indices, GL_STATIC_DRAW);

    glGenBuffers(1, &texture_coords_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, texture_coords_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(tex_coords), tex_coords, GL_STATIC_DRAW);
    a_texture_coords_loc = glGetAttribLocation(shader_program, "in_tex_coord");
    glVertexAttribPointer(a_texture_coords_loc, 2, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(a_texture_coords_loc);

    glBindVertexArray(0);

    return 1;
}

void update() {
    unsigned int time_now = SDL_GetTicks();
    if (last_time != 0) {
        unsigned long elapsed = time_now - last_time;

        x_angle += (0.7 * elapsed) / 1000.0;
        y_angle += (0.7 * elapsed) / 1000.0;
        z_angle += (0.7 * elapsed) / 1000.0;

        if (x_angle > 2*3.1415) x_angle -= 2*3.1415;
        if (y_angle > 2*3.1415) y_angle -= 2*3.1415;
        if (z_angle > 2*3.1415) z_angle -= 2*3.1415;
    }
    last_time = time_now;
}

void render() {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    get_identity_matrix(modelview_matrix);

    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -(near+far)/MAX_UNITS);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, 16 * sizeof(float));

    get_rotation_x_matrix4x4(aux_matrix1, x_angle);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    get_rotation_y_matrix4x4(aux_matrix1, y_angle);
    mult_matrices(modelview_matrix, aux_matrix2, aux_matrix1, 4);
    get_rotation_y_matrix4x4(aux_matrix1, z_angle);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, 16 * sizeof(float));

    mult_matrices(mvp_matrix, projection_matrix, modelview_matrix, 4);
    transpose_matrix(aux_matrix1, mvp_matrix, 4);

    memcpy(mvp_matrix, aux_matrix1, 16 * sizeof(float));



    glBindVertexArray(vertex_array);
    glUseProgram(shader_program);

    /*
    printf("------------------------\n");
    dump_matrix4x4(mvp_matrix);
    printf("------------------------\n");
    */
    glBindTexture(GL_TEXTURE_2D, texture);
    glUniform1i(glGetUniformLocation(shader_program, "tex"), 0);
    u_mvpMatrix_loc = glGetUniformLocation(shader_program, "u_mvp_matrix");
    glUniformMatrix4fv(u_mvpMatrix_loc, 1, GL_FALSE, mvp_matrix);

    glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_INT, NULL);

    glUseProgram(0);
    glBindVertexArray(0);
}

void cleanup() {
    if (vertex_indices_buffer) {
        glDeleteBuffers(1, &vertex_indices_buffer);
    }
    if (vertex_buffer) {
        glDeleteBuffers(1, &vertex_buffer);
    }
    if (vertex_array) {
        glDeleteVertexArrays(1, &vertex_array);
    }

    if (shader_program && fragment_shader) {
        glDetachShader(shader_program, fragment_shader);
    }
    if (shader_program && vertex_shader) {
        glDetachShader(shader_program, vertex_shader);
    }
    if (fragment_shader) {
        glDeleteShader(fragment_shader);
    }
    if (vertex_shader) {
        glDeleteShader(vertex_shader);
    }
    if (shader_program) {
        glDeleteProgram(shader_program);
    }

    glDeleteTextures(1, &texture);
    if (image2d) {
        SDL_FreeSurface(image2d);
        image2d = NULL;
    }

    if (gContext) {
        SDL_GL_DeleteContext(gContext);
        gContext = NULL;
    }

    if (gWindow) {
        SDL_DestroyWindow(gWindow);
        gWindow = NULL;
    }

    IMG_Quit();
    SDL_Quit();
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    quit = 0;
    while (!quit) {
        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
            }
        }

        update();
        render();

        SDL_GL_SwapWindow(gWindow);
    }

    cleanup();

    return 0;
}



char *text_file_read(char *fname) {
    FILE *fp;
    char *content = NULL;
    int count = 0;

    if (fname != NULL) {
        fp = fopen(fname, "rt");

        if (fp != NULL) {
            fseek(fp, 0, SEEK_END);
            count = ftell(fp);
            rewind(fp);

            if (count > 0) {
                content = (char*)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


void print_shader_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}

void print_program_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}


int check_gl_error() {
    GLenum error;

    error = glGetError();
    switch (error) {
        case GL_CONTEXT_LOST:
            printf("GL_CONTEXT_LOST [%d]: %s\n", error, gluErrorString(error));
            return GL_CONTEXT_LOST;
        case GL_INVALID_ENUM:
            printf("GL_INVALID_ENUM [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_ENUM;
        case GL_INVALID_VALUE:
            printf("GL_INVALID_VALUE [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_VALUE;
        case GL_INVALID_OPERATION:
            printf("GL_INVALID_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            printf("GL_INVALID_FRAMEBUFFER_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_FRAMEBUFFER_OPERATION;
        case GL_OUT_OF_MEMORY:
            printf("GL_OUT_OF_MEMORY [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_STACK_OVERFLOW:
            printf("GL_STACK_OVERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_OVERFLOW;
        case GL_STACK_UNDERFLOW:
            printf("GL_STACK_UNDERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_UNDERFLOW;
        case GL_NO_ERROR:
            printf("GL_NO_ERROR [%d]\n", error);
            break;
    }

    return GL_NO_ERROR;
}
